# Pfilter/Interfaces.pl
#
# library functions to handle network interfaces

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# list of functions in this file:

# function that when passed an address, returns which interface
# could include the address.
# sub address_to_interface($vars_ref, $address,
# 			   $error_lines_ref, $warning_lines_ref )

# function to return a list of all interface names that are listed
# sub default_route_interfaces( $vars_ref, 
# 			        $error_lines_ref, $warning_lines_ref )

# function to take two arrays of interface names, and return the 
# sub interfaces_intersection($vars_ref, 
# 			      $interfaces_1_ref, $interfaces_2_ref )

# set and clear a pair of flags on a network interface hash/structure
# sub interface_set_check_flags( $vars_ref, $interface_ref, 
#                                $set_flag, $set_text,
#                                $clear_flag, $clear_text, 
#                                $byonin, $modified_ref )

# function to return non-zero if a passed source address/range can
# sub valid_source_destination_combination( $vars_ref, 
# 					    $source, $destination, 
# 					    $error_lines_ref, 
# 					    $warning_lines_ref )

#=======================================================================

# function that when passed an address, returns which interface
# could include the address. If the address is not directly in 
# one or more of the interface address ranges, a zero length
# string is returned.

sub address_to_interface {
    my ($vars_ref, $address, $error_lines_ref, $warning_lines_ref ) = @_;

    # find a list of all the interfaces that match this address
    my @matching_interfaces = address_to_interfaces( $vars_ref,
						     $address, 
						     $error_lines_ref,
						     $warning_lines_ref );

    # there should normally be only one interface in the 
    # list, but if there are more than one interface, the
    # first is returned and an error message is pushed.
    my $result;
    # if no interface matched that address, return nothing
    if ( scalar @matching_interfaces == 0 ) {
	$result = "";
    }
    # if exactly one interface matched that address, return it
    elsif ( scalar @matching_interfaces == 1 ) {
	$result = $matching_interfaces[0];
    }
    # since more than one interface matched the address,
    # push an error message and return the first one
    else {
	$result = $matching_interfaces[0];
	my $string = join( " ", @matching_interfaces );
	push @$error_lines_ref, 
	"all of the ($string) interfaces matched address $address, using $result";
    }
    print "$0: exiting  address_to_interface address=$address result=$result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function that when passed an address, returns a list of the 
# interfaces that could include the address. If the address is 
# not directly in one or more of the interface address ranges, 
# a null list is returned.

sub address_to_interfaces {
    my ($vars_ref, $address, $error_lines_ref, $warning_lines_ref ) = @_;
#    print "$0: entering address_to_interfaces address=$address\n"
#	if ${$$vars_ref{constants}}{debug};

    my @matching_interfaces = ();

    # if the address is actually a range, let the range_to_interfaces
    # function handle this
    if ( $address =~ /\// ) {
	@matching_interfaces = range_to_interfaces( $vars_ref, 
						    $address, 
						    $error_lines_ref,
						    $warning_lines_ref );
	if ( ${$$vars_ref{constants}}{debug} ) {
            my $string = join( ',', @matching_interfaces );
	    print "$0: exiting  address_to_interfaces address=$address returning ($string)\n";
	}
	return @matching_interfaces;
    }

    # find the numeric address(es) that match this numeric/dns address
    my @ips = address_to_ips( $vars_ref, $address );
    if ( ! @ips ) {
	push @$error_lines_ref, "invalid address $address";
	print "$0: exiting  address_to_interfaces invalid address $address returning ()\n"
	    if ${$$vars_ref{constants}}{debug};
	return ();
    }

    # for each interface, ...
    my $interfaces_ref = $$vars_ref{interfaces};
    foreach my $interface_name ( sort ( keys %$interfaces_ref ) ) {
	my $interface_ref = $$interfaces_ref{$interface_name};
	# figure out the address range of that interface
	my $interface_range = ( exists $$interface_ref{mask} && 
				$$interface_ref{mask} ne "" ) ?
				"$$interface_ref{address}/$$interface_ref{mask}" :
				$$interface_ref{address};
	# for each resolved address, ...
	foreach my $ip ( @ips ) {
	    # if the input address is in this interface's
	    # address range, add the interface name to the
	    # list of matching interfaces
	    if ( address_in_range( $vars_ref,
				   $ip,
				   $interface_range,
				   $error_lines_ref,
				   $warning_lines_ref ) ) {
		push @matching_interfaces, $interface_name;
		last;
	    }
	}
    }
    # end of interfaces loop, the @matching_interfaces list
    # now contains all interfaces that include the passed address.
    # if no interfaces directly contained this address,
    # use the interfaces that are marked as default routes
    # (yes there should only be one interface marked as
    # a default route, but mistakes happen)
#    @matching_interfaces = default_route_interfaces( $vars_ref,
#						     $error_lines_ref,
#						     $warning_lines_ref )
    @matching_interfaces = @{$$vars_ref{default_route_interfaces}}
        if ! @matching_interfaces;

    if ( ${$$vars_ref{constants}}{debug} ) {
        my $string = join( ',', @matching_interfaces );
        print "$0: exiting  address_to_interfaces address=$address returning ($string)\n";
    }
    return @matching_interfaces;
}

#=======================================================================

# function to return a list of all interface names that are listed
# as default route gateways in the routing table.

sub default_route_interfaces {
    my ( $vars_ref, $error_lines_ref, $warning_lines_ref ) = @_;

    my $route_command = "route -n";
    my @route_output = `$route_command`;
    if ( ! ( @route_output ) ) {
	push @$error_lines_ref, 
	"cannot get route information with command <$route_command>";
	return 
    }
    chomp @route_output;
    my @default_route_lines = grep ( /^0\.0\.0\.0\s/, @route_output );
    my @interfaces;
    foreach my $line ( @default_route_lines ) {
	$line =~ s/^.*\s//;
	# should the following be ? split( ' ', $line);
	my @fields = split( /\s/, $line);
	push @interfaces, $fields[$#fields];
    }
    @interfaces = sort_uniq( @interfaces );
    if ( ${$$vars_ref{constants}}{debug} ) {
	my $string = join( ",", @interfaces );
	print "$0: exiting  default_route_interfaces returning ($string)\n" ;
    }
    return @interfaces;
}

#=======================================================================

# find all current real interfaces, returns reference to interfaces hash:
#   hash has interface names as keys, values are hash references,
#     referenced hashes include these key/value pairs:
#       address    interface IP address
#       broadcast  interface broadcast address
#       mask       interface IP mask
#       rfc1918    non-zero if true

sub find_interfaces {
    my ( $vars_ref ) = @_;

    my @ifconfig_output = `${$$vars_ref{defines}}{ifconfig_path} 2>/dev/null`;
    unless ( @ifconfig_output ) {
	die "$0: cannot find interfaces with ${$$vars_ref{constants}}{ifconfig_path} command\n"
	    if $$vars_ref{exec};
	print "$0: warning - cannot find interfaces with ${$$vars_ref{defines}}{ifconfig_path} command\n";
    }
    my $address;
    my $broadcast = "";
    my %interfaces;
    my $mask;
    my $name;
    my $ptp;
    foreach (@ifconfig_output) {
	$name = $1 if /^(\S+)/;
	next unless $name =~ /^[a-z]+\d*$/;
	next if $name eq "lo";
	next unless ( ($address, $broadcast, $mask) =
	    /^\s+inet addr:(\S+)\s+Bcast:(\S+)\s+Mask:(\S+)/ )
	    ||
	    ( ($address, $ptp, $mask) =
	    /^\s+inet addr:(\S+)\s+P-t-P:(\S+)\s+Mask:(\S+)/ );
	die "$0: invalid ip address $address found for interface $name\n"
	    unless valid_ip($vars_ref, $address);
#	print "$0: invalid broadcast address $broadcast found for interface $name\n"
#	    unless valid_bcast($vars_ref, $broadcast);
	die "$0: invalid ip mask $mask found for interface $name\n"
	    unless valid_mask($vars_ref, $mask);
	my %interface = (
			 address   => $address,
			 mask      => $mask,
			 name      => $name,  # needed so we can find name
			                      # if passed hash reference
			 rfc1918   => valid_rfc1918_ip($vars_ref, $address)
			 );
	$interface{broadcast} = $broadcast if $broadcast ne "";
	$interfaces{$name} = \%interface;
	$broadcast = "";
    }
    return \%interfaces;
}

#=======================================================================
#
# set and clear a pair of flags on a network interface hash/structure

sub interface_set_check_conflicting_flags {
    my ( $vars_ref, $interface_ref, $set_flag, $set_text,
	 $conflicting_flag, $conflicting_text, $byonin )
	= @_;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
#    if ( $debug ) {
#	print "$0: entering interface_set_check_conflicting"
#	    . " set flag=$set_flag text=$set_text"
#		. " conflicting flag=$conflicting_flag text=$conflicting_text"
#		    . " interface=$$interface_ref{name}\n";
#    }

    if ( exists $$interface_ref{$conflicting_flag} &&
	 $$interface_ref{$conflicting_flag} ) {
	print "$0: warning - interface $$interface_ref{name} was set as $conflicting_text,\n";
	print "    changed to $set_text $byonin\n";
	$$interface_ref{$conflicting_flag} = 0;
    }

    if ( ! exists $$interface_ref{$set_flag} ||
	 ! $$interface_ref{$set_flag} ) {
	print "$0: $$interface_ref{name} set $set_flag $byonin\n" 
	    if $debug;
	$$interface_ref{$set_flag} = 1;
    }
}

#=======================================================================
#
# function to take two arrays of interface names, and return the 
# intersection of those arrays, taking note of "ANY" in either

sub interfaces_intersection {
    my ($vars_ref, $interfaces_1_ref, $interfaces_2_ref ) = @_;
    my @intersecting_interfaces;
#    print "$0: entering interfaces_intersection interfaces= <@$interfaces_1_ref> <@$interfaces_2_ref>\n"
#	if ${$$vars_ref{constants}}{debug};

    # if both interfaces arrays include ANY, return ANY as the result
    if ( grep( /^ANY$/, @$interfaces_1_ref ) &&
	 grep( /^ANY$/, @$interfaces_2_ref ) ) {
	@intersecting_interfaces = ( "ANY" );

    # if one interfaces arrays includes ANY, return the other as the result
    } elsif (   grep( /^ANY$/, @$interfaces_1_ref ) &&
	      ! grep( /^ANY$/, @$interfaces_2_ref ) ) {
	@intersecting_interfaces = @$interfaces_2_ref;
    } elsif ( ! grep( /^ANY$/, @$interfaces_1_ref ) &&
		grep( /^ANY$/, @$interfaces_2_ref ) ) {
	@intersecting_interfaces = @$interfaces_1_ref;

    # otherwise, calculate the intersection
    } else {
	@intersecting_interfaces = ();
	foreach my $interface ( @$interfaces_1_ref ) {
	    push @intersecting_interfaces, $interface
		if grep( /^$interface$/, @$interfaces_2_ref);
	}
    }

    if ( ${$$vars_ref{constants}}{debug} ) {
        my $string_1 = join( ',', @$interfaces_1_ref );
        my $string_2 = join( ',', @$interfaces_2_ref );
        my $string_i = join( ',', @intersecting_interfaces );
        print "$0: exiting  interfaces_intersection input <$string_1> <$string_2> returning ($string_i)\n";
    }
    return @intersecting_interfaces;
}

#=======================================================================

# function that when passed a single address or an address range, 
# returns a list of which interfaces could include the single address
# or address range.

sub range_to_interfaces {
    my ($vars_ref, $range, $error_lines_ref, $warning_lines_ref ) = @_;
    my @matching_interfaces = ();
    print "$0: entering range_to_interfaces range=$range\n"
	if ${$$vars_ref{constants}}{debug};

    # if the input is a single address, use the simpler function
    if ( $range !~ /\// ) {
	@matching_interfaces = address_to_interfaces( $vars_ref,
						      $range,
						      $error_lines_ref,
						      $warning_lines_ref );
	if ( ${$$vars_ref{constants}}{debug} ) {
            my $string = join( ',', @matching_interfaces );
	    print "$0: exiting  range_to_interfaces address=$address returning ($string)\n";
	}
	return @matching_interfaces;
    }

    # resolve the range to a numeric ip range
    my $resolved_range;
    if ( ! range_to_ip_range( $vars_ref,
			      $range,
			      \$resolved_range,
			      $error_lines_ref,
			      $warning_lines_ref ) ) {
	push @$error_lines_ref, "invalid address/mask $range";
	return ();
    }

    # for each interface, ...
    my $interfaces_ref = $$vars_ref{interfaces};
    foreach my $interface_name ( keys %$interfaces_ref ) {
	my $interface_ref = $$interfaces_ref{$interface_name};

	# figure out the address range of that interface
	my $interface_range = $$interface_ref{address};
	$interface_range = "$interface_range\/$$interface_ref{mask}"
	    if exists $$interface_ref{mask} && 
	       $$interface_ref{mask} ne "";

        # see if the input resolved range intersects the 
        # range of the interface
        if ( range_intersects_range( $vars_ref,
				     $resolved_range,
				     $interface_range,
	                             $error_lines_ref,
				     $warning_lines_ref ) ) {
	    push @matching_interfaces, $interface_name;
	}
    }

    # end of interfaces loop, the @matching_interfaces list
    # now contains all interfaces that include the passed address.
    # if no interfaces directly contained this address,
    # use the interfaces that are marked as default routes
    # (yes there should only be one interface marked as
    # a default route, but mistakes happen)
#    @matching_interfaces = default_route_interfaces( $vars_ref,
#						     $error_lines_ref,
#						     $warning_lines_ref )

    @matching_interfaces = @{$$vars_ref{default_route_interfaces}}
        if ! @matching_interfaces;

    if ( ${$$vars_ref{constants}}{debug} ) {
        my $string = join( ',', @matching_interfaces );
        print "$0: exiting  range_to_interfaces range=$range returning ($string)\n";
    }
    return @matching_interfaces;
}

#=======================================================================

# return non-zero if a passed interface name is valid

sub valid_interface {
    my ($vars_ref, $interface) = @_;
    my $result;
    if ( grep ( /^$interface$/, @all_interfaces ) ) {
	$result = 1;
    } else {
	$result = return 0;
    }
    print "$0: exiting  valid_interface interface=$interface returning $result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================
#
# function to return non-zero if a passed source address/range can
# possibly be received by a given destination address/range.
# If the source address is "ANY" all filtered interfaces are used
# in the calculations. If the destination address is "ME" all
# filtered interfaces are used in the calculations.

sub valid_source_destination_combination {
    my ( $vars_ref, $source, $destination, 
	 $error_lines_ref, $warning_lines_ref ) = @_;
    my $result = 0;

    print "$0: entering valid_source_destination_combination source=$source dest=$destination\n"
	if ${$$vars_ref{constants}}{debug};

    # find all the interfaces that could match source address
    my @source_interfaces;
    if ( $source eq "ANY" ) {
	@source_interfaces = 
	    @{$$vars_ref{interface_names}};
#	    @{$$vars_ref{filtered_interface_names}};
    } else {
        @source_interfaces = 
	    address_to_interfaces( $vars_ref,
				   $source,
				   $error_lines_ref,
				   $warning_lines_ref );
    }
    if ( ${$$vars_ref{constants}}{debug} ) {
	my $string = join( ',', @source_interfaces );
	print "$0: in       valid_source_destination_combination source ints=($string)\n";
    }

    # find all the interfaces that could match destination address.
    # one tricky thing is that a destination address that is the
    # target of an alias, could also come into the faked pseudo
    # address of the alias.
    my @destination_interfaces;
    if ( $destination eq "ME" ) {
	@destination_interfaces = 
	    @{$$vars_ref{filtered_interface_names}};
    } else {
	@destination_interfaces = 
	    address_to_interfaces( $vars_ref,
				   $destination,
				   $error_lines_ref,
				   $warning_lines_ref );
	my $alias_real_to_faked_ref = $$vars_ref{alias_real_to_faked};
	if ( exists $$alias_real_to_faked_ref{$destination} ) {
#	    print_hash( "", 
#			"in       valid_source_destination_combination alias_real_to_faked",
#			$alias_real_to_faked_ref )
#		if ${$$vars_ref{constants}}{debug};
	    my $pseudo_address = 
		$$alias_real_to_faked_ref{$destination};
#	    print( "$0: in       valid_source_destination_combination pseudo=$pseudo_address\n")
#		if ${$$vars_ref{constants}}{debug};
	    my @pseudo_interfaces =
		address_to_interfaces( $vars_ref,
				       $pseudo_address,
				       $error_lines_ref,
				       $warning_lines_ref );
	    @destination_interfaces = 
                ( @destination_interfaces, @pseudo_interfaces );
#	    print( "$0: in       valid_source_destination_combination ints=<@destination_interfaces>\n")
#		if ${$$vars_ref{constants}}{debug};
	    @destination_interfaces = sort_uniq( @destination_interfaces );
	    print( "$0: in       valid_source_destination_combination ints=<@destination_interfaces>\n")
		if ${$$vars_ref{constants}}{debug};
	}
    }
    if ( ${$$vars_ref{constants}}{debug} ) {
	my $string = join( ',', @destination_interfaces );
	print "$0: in       valid_source_destination_combination dest ints=($string)\n";
    }

    # see if there are any common interfaces
    foreach my $source_interface ( @source_interfaces ) {
	$result = 1
	    if grep( /^$source_interface$/, @destination_interfaces );
    }

    print "$0: exiting  valid_source_destination_combination source=$source dest=$destination result =$result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;

