# Pfilter/Mining.pm
#
# common library functions for mining data from other pfilter data

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# function to mine network srevices from constants, defines, and macros
# sub mine_network_services( $vars_ref, $error_lines_ref, $warning_lines_ref )

# function to process a mined network service from a constant or define
# sub process_simple_service( $vars_ref, $service_name, 
# 			      $service_protocols_ports, $source,
# 			      $error_lines_ref, $warning_lines_ref )

#=======================================================================

# function to mine network srevices from constants, defines, and macros
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
#     macros -> reference to %MACRO names/references hash
#   reference to error lines array
#   reference to warning lines array
# This function modifies the hash of all flags/values, with new/modified:
#   named_services_protocols_ports -> deletes replaced services,
#                       if the new network service definition is simply a 
#                       protocols/ports list the new service definition
#                       is inserted into this hash
#   network_services -> reference to new hash of new network services,
#                       each element key is the network service name,
#                       and each element value is a reference to a hash
#                       for that network service that has these values:
#     open              reference to text lines for OPEN service
#     close             reference to text lines for CLOSE service
#
sub mine_network_services
{
    my( $vars_ref, $error_lines_ref, $warning_lines_ref ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};
    print "$0: in mine_network_services\n" if $debug;

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};
    my $macros_ref = $$vars_ref{macros};

    # make new network services hash and link it in
    my %network_services;
    $$vars_ref{network_services} = \%network_services;

    # loop through the stored constants whose names
    # are of the form service-NAME-protocols-ports
    my @all_keys = keys %$constants_ref;
    my @service_names = grep ( /^service\-.+\-protocols\-ports$/, @all_keys );
    foreach my $service_key ( @service_names ) {
	my $service_name = $service_key;
	$service_name =~ s/^service\-//;
	$service_name =~ s/\-protocols\-ports$//;
        process_simple_service
	    ( $vars_ref, $service_name, $$constants_ref{$service_key},
	      $error_lines_ref, $warning_lines_ref );
    }

    # loop through the stored defines whose names
    # are of the form service-NAME-protocols-ports
    @all_keys = keys %$defines_ref;
    @service_names = grep ( /^service\-.+\-protocols\-ports$/, @all_keys );
    foreach my $service_key ( @service_names ) {
	my $service_name = $service_key;
	$service_name =~ s/^service\-//;
	$service_name =~ s/\-protocols\-ports$//;
        process_simple_service
	    ( $vars_ref, $service_name, $$defines_ref{$service_key},
	      $error_lines_ref, $warning_lines_ref );
    }

    # loop through the stored macros looking for pairs
    # named service-NAME-open and service-NAME-close
    # which will be the opening and closing macro scripts
    # for a network service.  first find all the service
    # names implied by all the service-SERVICE-open or 
    # service-SERVICE-close macros
    @all_keys = keys %$macros_ref;
    my @service_keys = grep ( /^service\-.+\-(open|close)$/, @all_keys );
    my %services;
    foreach my $service_key ( @service_keys ) {
	my $service = $service_key;
	$service =~ s/^service\-//;
	$service =~ s/\-(open|close)$//;
	$services{$service} = 1;
    }
    print "$0: in mine_services servicekeys=" .
        join(',', @service_keys) . "\n"
        if $debug;
    foreach my $service ( keys %services ) {
	print "$0: in mine_services defining service $service\n"
	    if $debug;
	my %service;
	my $success = 0;
	# if the service-NAME-open macro is defined, ...
	if ( exists $$macros_ref{"service-$service-open"} ) {
	    # if the service-NAME-open was there 
	    # but no service-NAME-close was there,
	    # issue a warning about that
	    # (but we will still store the lopsided service)
	    if ( ! exists $$macros_ref{"service-$service-close"} ) {
		push @warning_lines_ref,
	        "macro service-$service-open is defined " .
		"which implies a network service named $service, " .
		"but no service-$service-close macro is defined";
	    }
	    # figure out for the open part of this service:
	    # the macro lines to be inserted for the open,
	    # the names of the arguments for the open,
	    # and how many open arguments there are
	    my $open_macro_ref = $$macros_ref{"service-$service-open"};
	    my $open_arguments_ref = $$open_macro_ref{arguments};
	    my $open_arguments_num = scalar @$open_arguments_ref;
	    # if there are a wrong number of open arguments, ...
	    if ( $open_arguments_num != 6 ) {
		push @$error_lines_ref, 
		"network service $service being defined by macros has the wrong has the wrong number of arguments ($open_arguments_num instead of 6) in macro service-$service-open"
	    # if the right number of open arguments are present, 
	    # store everything we need for the open portion
	    } else {
		$service{open} = "service-$service-open";
		$success = 1;
	    }
	} # end if the service-NAME-open macro was defined
	# take care of the open portion of the service
	if ( exists $$macros_ref{"service-$service-open"} ) {
	    # if the service-NAME-open was there 
	    # but no service-NAME-close was there,
	    # issue a warning about that
	    # (but we will still store the lopsided service)
	    if ( ! exists $$macros_ref{"service-$service-close"} ) {
		push @warning_lines_ref,
	        "macro service-$service-open is defined " .
		"which implies a network service named $service, " .
		"but no service-$service-close macro is defined";
	    }
	    # figure out for the open part of this service:
	    # the macro lines to be inserted for the open,
	    # the names of the arguments for the open,
	    # and how many open arguments there are
	    my $open_macro_ref = $$macros_ref{"service-$service-open"};
	    my $open_arguments_ref = $$open_macro_ref{arguments};
	    my $open_arguments_num = scalar @$open_arguments_ref;
	    # if there are a wrong number of open arguments, ...
	    if ( $open_arguments_num != 6 ) {
		push @$error_lines_ref, 
		"network service $service being defined by macros has the wrong has the wrong number of arguments ($open_arguments_num instead of 6) in macro service-$service-open"
	    # if the right number of open arguments are present, 
	    # store everything we need for the open portion
	    } else {
		$service{open} = "service-$service-open";
		$success = 1;
	    }
	} # end if the service-NAME-close macro was defined
	# take care of the close portion of the service
	if ( exists $$macros_ref{"service-$service-close"} ) {
	    # if the service-NAME-close was there 
	    # but no service-NAME-open was there,
	    # issue a warning about that
	    # (but we will still store the lopsided service)
	    if ( ! exists $$macros_ref{"service-$service-open"} ) {
		push @warning_lines_ref,
	        "macro service-$service-close is defined " .
		"which implies a network service named $service, " .
		"but no service-$service-open macro is defined";
	    }
	    # figure out for the close part of this service:
	    # the macro lines to be inserted for the close,
	    # the names of the arguments for the close,
	    # and how many close arguments there are
	    my $close_macro_ref = $$macros_ref{"service-$service-close"};
	    my $close_arguments_ref = $$close_macro_ref{arguments};
	    my $close_arguments_num = scalar @$close_arguments_ref;
	    # if there are a wrong number of close arguments, ...
	    if ( $close_arguments_num != 6 ) {
		push @$error_lines_ref, 
		"network service $service being defined by macros has the wrong has the wrong number of arguments ($close_arguments_num instead of 6) in macro service-$service-close"
	    # if the right number of close arguments are present, 
	    # store everything we need for the close portion
	    } else {
		$service{close} = "service-$service-close";
		$success = 1;
	    }
	} # end if the service-NAME-close macro was defined
	# if either the open or close portion of the service was 
	# successfully stored, save the service hash in the 
	# services hash
	print_hash( "", "defined service $service", \%service ) if $debug;
	$services{$service} = \%service if $success;
    } # end for each service name

    # save named services in common variables space
    $$vars_ref{services} = \%services;

#    print_hash( "", "exiting mine_network_services named_services_protocols_ports",
#		$$vars_ref{named_services_protocols_ports} )
#        if $debug;
    print_hash( "", "exiting mine_network_services network_services",
		$$vars_ref{services} )
        if $debug;
}

#=======================================================================

# function to process a mined network service from a constant or define
#
# Passed parameters are:
#   reference to hash of all flags/values of interest
#   found network service name string
#   found network service protocols/ports combined string
#   "constant" or "define" depending on where this came from
#   reference to error lines array
#   reference to warning lines array
# This function modifies the hash of all flags/values, with new/modified:
#   named_services_protocols_ports -> deletes replaced services,
#                       if the new network service definition is simply a 
#                       protocols/ports list the new service definition
#                       is inserted into this hash
#
sub process_simple_service
{
    my( $vars_ref, $service_name, $service_protocols_ports, $source,
	$error_lines_ref, $warning_lines_ref ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};
    my $named_services_protocols_ports_ref = 
        $$vars_ref{named_services_protocols_ports};
    my $protocols_ref = $$vars_ref{protocols};

    # parse out the protocols(s) and port(s)
    my @slash_fields = split( /\//, $service_protocols_ports );
    if ( scalar @slash_fields != 2 ) {
	push @$error_lines_ref,
	"cannot use service-$source-protocols-ports as network service definition illegal value <$service_protocols_ports>";
	return;
    }
    my @service_protocols = split( /\s/, $slash_fields[0] );
    my @service_ports = split( /\s/, $slash_fields[1] );
    if ( scalar @service_protocols < 1 || scalar @service_ports < 1 ) {
	push @$error_lines_ref,
	"cannot use service-$source-protocols-ports as network service definition illegal value <$service_protocols_ports>";
	return;
    }
    print "$0: in process_simple_service name=$service_name protocols=(".
        join(',',@service_protocols) . ") ports=(" . join(',',@service_ports) . ")\n" 
        if $debug;
    # warn if using unknown protocol names, but use them anyway since
    # the programs that will be called to insert the filtering rules
    # might know of protocol names that we do not know
    foreach my $protocol ( @service_protocols ) {
	if ( ! exists $$protocols_ref{$protocol} 
	     && $protocol !~ /^[0-9]+$/ ) {
	    push @warning_lines_ref,
	    "newly (re)defined network service $service_name uses unknown protocol $protocol";
	}
    }
    # it's an error if a ports value isn't either a simple number,
    # or a number range
    foreach my $port ( @service_ports ) {
	if ( $port !~ /^[0-9]+$/ 
	     && $port !~ /^[0-9]+\:[0-9]+$/ ) {
	    push @error_lines_ref,
	    "newly (re)defined network service $service_name uses bad port(s) value $port";
	    return;
	}
    }

    # get rid of a previous definition if one was there from /etc/services
    if ( exists $$named_services_protocols_ports_ref{$service_name} ) {
	print "$0: in process_simple_service replacing named service $service_name\n"
	    if $debug;
	delete $$named_services_protocols_ports_ref{$service_name};
    }

    # add the new definition
    my %hash_ports;
    foreach my $port ( @service_ports ) {
        $hash_ports{$port} = 1;
    }
    my %hash_protocols_ports;
    foreach my $protocol ( @service_protocols ) {
	$hash_protocols_ports{$protocol} = \%hash_ports;
    }
    print_hash( "", "in process_simple_service adding named service $service_name",
		\%hash_protocols_ports )
        if $debug;
    $$named_services_protocols_ports_ref{$service_name} = \%hash_protocols_ports;
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;

